####################################################
# File is managed by puppet                        #
# Module: anycast_healthchecker                    #
# Path:   templates/anycast-healthchecker.conf.erb #
####################################################
[DEFAULT]
interface            = lo

[daemon]
pidfile              = <%= @pidfile %>
bird_conf            = <%= @bird_conf %>
bird_variable        = <%= @bird_variable %>
bird_reconfigure_cmd = <%= @bird_reconfigure_cmd %>
bird_keep_changes    = <%= @bird_keep_changes %>
bird_changes_counter = <%= @bird_changes_counter %>
loglevel             = <%= @log_level %>
log_maxbytes         = <%= @log_maxbytes %>
log_backups          = <%= @log_backups %>
log_file             = <%= @log_file %>
stderr_file          = <%= @stderr_file %>
stdout_file          = <%= @stdout_file %>
dummy_ip_prefix      = <%= @dummy_ip_prefix %>
<%- if @ipv6 -%>
ipv6                  = <%= @ipv6 %>
bird6_conf            = <%= @bird6_conf %>
bird6_variable        = <%= @bird6_variable %>
bird6_reconfigure_cmd = <%= @bird6_reconfigure_cmd %>
bird6_keep_changes    = <%= @bird6_keep_changes %>
bird6_changes_counter = <%= @bird6_changes_counter %>
dummy_ip6_prefix      = <%= @dummy_ip6_prefix %>
<%- end -%>
<%- if @json_logging -%>
json_logging         = true
http_server          = <%= @http_server %>
http_server_protocol = <%= @http_server_protocol %>
http_server_port     = <%= @http_server_port %>
http_server_timeout  = <%= @http_server_timeout -%>
<%- end %>
